@@ -49,7 +49,6 @@
 #include "clutter-event-private.h"
 #include "clutter-marshal.h"
 #include "clutter-private.h"
-#include "clutter-profile.h"
 #include "clutter-stage-manager-private.h"
 #include "clutter-stage-private.h"
 #include "clutter-stage-window.h"
@@ -58,9 +57,9 @@
 #define CLUTTER_DISABLE_DEPRECATION_WARNINGS
 #include "deprecated/clutter-backend.h"
 
-#ifdef HAVE_CLUTTER_WAYLAND_COMPOSITOR
+#ifdef CLUTTER_HAS_WAYLAND_COMPOSITOR_SUPPORT
 #include "wayland/clutter-wayland-compositor.h"
-#endif /* HAVE_CLUTTER_WAYLAND_COMPOSITOR */
+#endif
 
 #include <cogl/cogl.h>
 
@@ -89,8 +88,14 @@
 #ifdef CLUTTER_INPUT_WAYLAND
 #include "wayland/clutter-device-manager-wayland.h"
 #endif
+#ifdef CLUTTER_WINDOWING_MIR
+#include "mir/clutter-backend-mir.h"
+#endif
+#ifdef CLUTTER_INPUT_MIR
+#include "mir/clutter-device-manager-mir.h"
+#endif
 
-#ifdef HAVE_CLUTTER_WAYLAND_COMPOSITOR
+#ifdef CLUTTER_HAS_WAYLAND_COMPOSITOR_SUPPORT
 #include <cogl/cogl-wayland-server.h>
 #include <wayland-server.h>
 #include "wayland/clutter-wayland-compositor.h"
@@ -98,18 +103,6 @@
 
 #define DEFAULT_FONT_NAME       "Sans 10"
 
-struct _ClutterBackendPrivate
-{
-  cairo_font_options_t *font_options;
-
-  gchar *font_name;
-
-  gfloat units_per_em;
-  gint32 units_serial;
-
-  GList *event_translators;
-};
-
 enum
 {
   RESOLUTION_CHANGED,
@@ -119,13 +112,13 @@ enum
   LAST_SIGNAL
 };
 
-G_DEFINE_ABSTRACT_TYPE_WITH_PRIVATE (ClutterBackend, clutter_backend, G_TYPE_OBJECT)
+G_DEFINE_ABSTRACT_TYPE (ClutterBackend, clutter_backend, G_TYPE_OBJECT)
 
 static guint backend_signals[LAST_SIGNAL] = { 0, };
 
 /* Global for being able to specify a compositor side wayland display
  * pointer before clutter initialization */
-#ifdef HAVE_CLUTTER_WAYLAND_COMPOSITOR
+#ifdef CLUTTER_HAS_WAYLAND_COMPOSITOR_SUPPORT
 static struct wl_display *_wayland_compositor_display;
 #endif
 
@@ -134,17 +127,15 @@ static const char *allowed_backend;
 static void
 clutter_backend_dispose (GObject *gobject)
 {
-  ClutterBackendPrivate *priv = CLUTTER_BACKEND (gobject)->priv;
+  ClutterBackend *backend = CLUTTER_BACKEND (gobject);
 
   /* clear the events still in the queue of the main context */
   _clutter_clear_events_queue ();
 
   /* remove all event translators */
-  if (priv->event_translators != NULL)
-    {
-      g_list_free (priv->event_translators);
-      priv->event_translators = NULL;
-    }
+  g_clear_pointer (&backend->event_translators, g_list_free);
+
+  g_clear_pointer (&backend->dummy_onscreen, cogl_object_unref);
 
   G_OBJECT_CLASS (clutter_backend_parent_class)->dispose (gobject);
 }
@@ -156,7 +147,7 @@ clutter_backend_finalize (GObject *gobject)
 
   g_source_destroy (backend->cogl_source);
 
-  g_free (backend->priv->font_name);
+  g_free (backend->font_name);
   clutter_backend_set_font_options (backend, NULL);
 
   G_OBJECT_CLASS (clutter_backend_parent_class)->finalize (gobject);
@@ -221,7 +212,6 @@ get_units_per_em (ClutterBackend       *backend,
 static void
 clutter_backend_real_resolution_changed (ClutterBackend *backend)
 {
-  ClutterBackendPrivate *priv = backend->priv;
   ClutterMainContext *context;
   ClutterSettings *settings;
   gdouble resolution;
@@ -239,26 +229,25 @@ clutter_backend_real_resolution_changed (ClutterBackend *backend)
   if (context->font_map != NULL)
     cogl_pango_font_map_set_resolution (context->font_map, resolution);
 
-  priv->units_per_em = get_units_per_em (backend, NULL);
-  priv->units_serial += 1;
+  backend->units_per_em = get_units_per_em (backend, NULL);
+  backend->units_serial += 1;
 
-  CLUTTER_NOTE (BACKEND, "Units per em: %.2f", priv->units_per_em);
+  CLUTTER_NOTE (BACKEND, "Units per em: %.2f", backend->units_per_em);
 }
 
 static void
 clutter_backend_real_font_changed (ClutterBackend *backend)
 {
-  ClutterBackendPrivate *priv = backend->priv;
+  backend->units_per_em = get_units_per_em (backend, NULL);
+  backend->units_serial += 1;
 
-  priv->units_per_em = get_units_per_em (backend, NULL);
-  priv->units_serial += 1;
-
-  CLUTTER_NOTE (BACKEND, "Units per em: %.2f", priv->units_per_em);
+  CLUTTER_NOTE (BACKEND, "Units per em: %.2f", backend->units_per_em);
 }
 
 static gboolean
-clutter_backend_real_create_context (ClutterBackend  *backend,
-                                     GError         **error)
+clutter_backend_do_real_create_context (ClutterBackend  *backend,
+                                        CoglDriver       driver_id,
+                                        GError         **error)
 {
   ClutterBackendClass *klass;
   CoglSwapChain *swap_chain;
@@ -281,7 +270,7 @@ clutter_backend_real_create_context (ClutterBackend  *backend,
   if (backend->cogl_renderer == NULL)
     goto error;
 
-#ifdef HAVE_CLUTTER_WAYLAND_COMPOSITOR
+#ifdef CLUTTER_HAS_WAYLAND_COMPOSITOR_SUPPORT
   /* If the application is trying to act as a Wayland compositor then
      it needs to have an EGL-based renderer backend */
   if (_wayland_compositor_display)
@@ -290,6 +279,7 @@ clutter_backend_real_create_context (ClutterBackend  *backend,
 #endif
 
   CLUTTER_NOTE (BACKEND, "Connecting the renderer");
+  cogl_renderer_set_driver (backend->cogl_renderer, driver_id);
   if (!cogl_renderer_connect (backend->cogl_renderer, &internal_error))
     goto error;
 
@@ -332,7 +322,7 @@ clutter_backend_real_create_context (ClutterBackend  *backend,
   if (backend->cogl_display == NULL)
     goto error;
 
-#ifdef HAVE_CLUTTER_WAYLAND_COMPOSITOR
+#ifdef CLUTTER_HAS_WAYLAND_COMPOSITOR_SUPPORT
   cogl_wayland_display_set_compositor_display (backend->cogl_display,
                                                _wayland_compositor_display);
 #endif
@@ -346,10 +336,6 @@ clutter_backend_real_create_context (ClutterBackend  *backend,
   if (backend->cogl_context == NULL)
     goto error;
 
-  backend->cogl_source = cogl_glib_source_new (backend->cogl_context,
-                                               G_PRIORITY_DEFAULT);
-  g_source_attach (backend->cogl_source, NULL);
-
   /* the display owns the renderer and the swap chain */
   cogl_object_unref (backend->cogl_renderer);
   cogl_object_unref (swap_chain);
@@ -372,16 +358,61 @@ error:
   if (swap_chain != NULL)
     cogl_object_unref (swap_chain);
 
-  if (internal_error != NULL)
-    g_propagate_error (error, internal_error);
-  else
-    g_set_error_literal (error, CLUTTER_INIT_ERROR,
-                         CLUTTER_INIT_ERROR_BACKEND,
-                         _("Unable to initialize the Clutter backend"));
-
   return FALSE;
 }
 
+static gboolean
+clutter_backend_real_create_context (ClutterBackend  *backend,
+                                     GError         **error)
+{
+  static const struct {
+    const char *driver_name;
+    CoglDriver driver_id;
+  } known_drivers[] = {
+    { "GL3", COGL_DRIVER_GL3 },
+    { "GL (Legacy)", COGL_DRIVER_GL },
+    { "GLES 2.0", COGL_DRIVER_GLES2 },
+    { "ANY", COGL_DRIVER_ANY },
+  };
+
+  GError *internal_error = NULL;
+  int i;
+
+  for (i = 0; i < G_N_ELEMENTS (known_drivers); i++)
+    {
+      CLUTTER_NOTE (BACKEND, "Checking for the %s driver", known_drivers[i].driver_name);
+
+      if (clutter_backend_do_real_create_context (backend, known_drivers[i].driver_id, &internal_error))
+        break;
+
+      if (internal_error)
+        {
+          CLUTTER_NOTE (BACKEND, "Unable to use the %s driver: %s",
+                        known_drivers[i].driver_name,
+                        internal_error->message);
+          g_clear_error (&internal_error);
+        }
+    }
+
+  if (backend->cogl_context == NULL)
+    {
+      if (internal_error != NULL)
+        g_propagate_error (error, internal_error);
+      else
+        g_set_error_literal (error, CLUTTER_INIT_ERROR,
+                             CLUTTER_INIT_ERROR_BACKEND,
+                            _("Unable to initialize the Clutter backend"));
+
+      return FALSE;
+    }
+
+  backend->cogl_source = cogl_glib_source_new (backend->cogl_context,
+                                               G_PRIORITY_DEFAULT);
+  g_source_attach (backend->cogl_source, NULL);
+
+  return TRUE;
+}
+
 static void
 clutter_backend_real_ensure_context (ClutterBackend *backend,
                                      ClutterStage   *stage)
@@ -491,6 +522,11 @@ _clutter_create_backend (void)
     retval = g_object_new (CLUTTER_TYPE_BACKEND_WIN32, NULL);
   else
 #endif
+#ifdef CLUTTER_WINDOWING_GDK
+  if (backend == NULL || backend == I_(CLUTTER_WINDOWING_GDK))
+    retval = g_object_new (CLUTTER_TYPE_BACKEND_GDK, NULL);
+  else
+#endif
 #ifdef CLUTTER_WINDOWING_X11
   if (backend == NULL || backend == I_(CLUTTER_WINDOWING_X11))
     retval = g_object_new (CLUTTER_TYPE_BACKEND_X11, NULL);
@@ -506,9 +542,9 @@ _clutter_create_backend (void)
     retval = g_object_new (CLUTTER_TYPE_BACKEND_EGL_NATIVE, NULL);
   else
 #endif
-#ifdef CLUTTER_WINDOWING_GDK
-  if (backend == NULL || backend == I_(CLUTTER_WINDOWING_GDK))
-    retval = g_object_new (CLUTTER_TYPE_BACKEND_GDK, NULL);
+#ifdef CLUTTER_WINDOWING_MIR
+  if (backend == NULL || backend == I_(CLUTTER_WINDOWING_MIR))
+    retval = g_object_new (CLUTTER_TYPE_BACKEND_MIR, NULL);
   else
 #endif
   if (backend == NULL)
@@ -589,6 +625,14 @@ clutter_backend_real_init_events (ClutterBackend *backend)
       _clutter_events_wayland_init (backend);
     }
   else
+#endif
+#ifdef CLUTTER_INPUT_MIR
+  if (clutter_check_windowing_backend (CLUTTER_WINDOWING_MIR) &&
+      (input_backend == NULL || input_backend == I_(CLUTTER_INPUT_MIR)))
+    {
+      _clutter_events_mir_init (backend);
+    }
+  else
 #endif
   if (input_backend != NULL)
     {
@@ -616,10 +660,9 @@ clutter_backend_real_translate_event (ClutterBackend *backend,
                                       gpointer        native,
                                       ClutterEvent   *event)
 {
-  ClutterBackendPrivate *priv = backend->priv;
   GList *l;
 
-  for (l = priv->event_translators;
+  for (l = backend->event_translators;
        l != NULL;
        l = l->next)
     {
@@ -719,9 +762,10 @@ clutter_backend_class_init (ClutterBackendClass *klass)
 static void
 clutter_backend_init (ClutterBackend *self)
 {
-  self->priv = clutter_backend_get_instance_private (self);
-  self->priv->units_per_em = -1.0;
-  self->priv->units_serial = 1;
+  self->units_per_em = -1.0;
+  self->units_serial = 1;
+
+  self->dummy_onscreen = COGL_INVALID_HANDLE;
 }
 
 void
@@ -821,11 +865,12 @@ _clutter_backend_ensure_context (ClutterBackend *backend,
   g_assert (CLUTTER_IS_BACKEND (backend));
   g_assert (CLUTTER_IS_STAGE (stage));
 
-  if (current_context_stage != stage || !CLUTTER_ACTOR_IS_REALIZED (stage))
+  if (current_context_stage != stage ||
+      !clutter_actor_is_realized (CLUTTER_ACTOR (stage)))
     {
       ClutterStage *new_stage = NULL;
 
-      if (!CLUTTER_ACTOR_IS_REALIZED (stage))
+      if (!clutter_actor_is_realized (CLUTTER_ACTOR (stage)))
         {
           new_stage = NULL;
 
@@ -938,18 +983,14 @@ gfloat
 _clutter_backend_get_units_per_em (ClutterBackend       *backend,
                                    PangoFontDescription *font_desc)
 {
-  ClutterBackendPrivate *priv;
-
-  priv = backend->priv;
-
   /* recompute for the font description, but do not cache the result */
   if (font_desc != NULL)
     return get_units_per_em (backend, font_desc);
 
-  if (priv->units_per_em < 0)
-    priv->units_per_em = get_units_per_em (backend, NULL);
+  if (backend->units_per_em < 0)
+    backend->units_per_em = get_units_per_em (backend, NULL);
 
-  return priv->units_per_em;
+  return backend->units_per_em;
 }
 
 void
@@ -1178,21 +1219,17 @@ void
 clutter_backend_set_font_options (ClutterBackend             *backend,
                                   const cairo_font_options_t *options)
 {
-  ClutterBackendPrivate *priv;
-
   g_return_if_fail (CLUTTER_IS_BACKEND (backend));
 
-  priv = backend->priv;
-
-  if (priv->font_options != options)
+  if (backend->font_options != options)
     {
-      if (priv->font_options)
-        cairo_font_options_destroy (priv->font_options);
+      if (backend->font_options)
+        cairo_font_options_destroy (backend->font_options);
 
       if (options)
-        priv->font_options = cairo_font_options_copy (options);
+        backend->font_options = cairo_font_options_copy (options);
       else
-        priv->font_options = NULL;
+        backend->font_options = NULL;
 
       g_signal_emit (backend, backend_signals[FONT_CHANGED], 0);
     }
@@ -1213,27 +1250,20 @@ clutter_backend_set_font_options (ClutterBackend             *backend,
 const cairo_font_options_t *
 clutter_backend_get_font_options (ClutterBackend *backend)
 {
-  ClutterBackendPrivate *priv;
-
   g_return_val_if_fail (CLUTTER_IS_BACKEND (backend), NULL);
 
-  priv = backend->priv;
+  if (G_LIKELY (backend->font_options))
+    return backend->font_options;
 
-  if (G_LIKELY (priv->font_options))
-    return priv->font_options;
+  backend->font_options = cairo_font_options_create ();
 
-  priv->font_options = cairo_font_options_create ();
-
-  cairo_font_options_set_hint_style (priv->font_options,
-                                     CAIRO_HINT_STYLE_NONE);
-  cairo_font_options_set_subpixel_order (priv->font_options,
-                                         CAIRO_SUBPIXEL_ORDER_DEFAULT);
-  cairo_font_options_set_antialias (priv->font_options,
-                                    CAIRO_ANTIALIAS_DEFAULT);
+  cairo_font_options_set_hint_style (backend->font_options, CAIRO_HINT_STYLE_NONE);
+  cairo_font_options_set_subpixel_order (backend->font_options, CAIRO_SUBPIXEL_ORDER_DEFAULT);
+  cairo_font_options_set_antialias (backend->font_options, CAIRO_ANTIALIAS_DEFAULT);
 
   g_signal_emit (backend, backend_signals[FONT_CHANGED], 0);
 
-  return priv->font_options;
+  return backend->font_options;
 }
 
 /**
@@ -1276,28 +1306,25 @@ clutter_backend_set_font_name (ClutterBackend *backend,
 const gchar *
 clutter_backend_get_font_name (ClutterBackend *backend)
 {
-  ClutterBackendPrivate *priv;
   ClutterSettings *settings;
 
   g_return_val_if_fail (CLUTTER_IS_BACKEND (backend), NULL);
 
-  priv = backend->priv;
-
   settings = clutter_settings_get_default ();
 
   /* XXX yuck. but we return a const pointer, so we need to
    * store it in the backend
    */
-  g_free (priv->font_name);
-  g_object_get (settings, "font-name", &priv->font_name, NULL);
+  g_free (backend->font_name);
+  g_object_get (settings, "font-name", &backend->font_name, NULL);
 
-  return priv->font_name;
+  return backend->font_name;
 }
 
 gint32
 _clutter_backend_get_units_serial (ClutterBackend *backend)
 {
-  return backend->priv->units_serial;
+  return backend->units_serial;
 }
 
 gboolean
@@ -1314,26 +1341,22 @@ void
 _clutter_backend_add_event_translator (ClutterBackend         *backend,
                                        ClutterEventTranslator *translator)
 {
-  ClutterBackendPrivate *priv = backend->priv;
-
-  if (g_list_find (priv->event_translators, translator) != NULL)
+  if (g_list_find (backend->event_translators, translator) != NULL)
     return;
 
-  priv->event_translators =
-    g_list_prepend (priv->event_translators, translator);
+  backend->event_translators =
+    g_list_prepend (backend->event_translators, translator);
 }
 
 void
 _clutter_backend_remove_event_translator (ClutterBackend         *backend,
                                           ClutterEventTranslator *translator)
 {
-  ClutterBackendPrivate *priv = backend->priv;
-
-  if (g_list_find (priv->event_translators, translator) == NULL)
+  if (g_list_find (backend->event_translators, translator) == NULL)
     return;
 
-  priv->event_translators =
-    g_list_remove (priv->event_translators, translator);
+  backend->event_translators =
+    g_list_remove (backend->event_translators, translator);
 }
 
 /**
@@ -1344,12 +1367,12 @@ _clutter_backend_remove_event_translator (ClutterBackend         *backend,
  * @backend. A #CoglContext is required when using some of the
  * experimental 2.0 Cogl API.
  *
- * <note>Since CoglContext is itself experimental API this API should
- * be considered experimental too.</note>
+ * Since CoglContext is itself experimental API this API should
+ * be considered experimental too.
  *
- * <note>This API is not yet supported on OSX because OSX still
+ * This API is not yet supported on OSX because OSX still
  * uses the stub Cogl winsys and the Clutter backend doesn't
- * explicitly create a CoglContext.</note>
+ * explicitly create a CoglContext.
  *
  * Return value: (transfer none): The #CoglContext associated with @backend.
  *
@@ -1362,7 +1385,7 @@ clutter_backend_get_cogl_context (ClutterBackend *backend)
   return backend->cogl_context;
 }
 
-#ifdef HAVE_CLUTTER_WAYLAND_COMPOSITOR
+#ifdef CLUTTER_HAS_WAYLAND_COMPOSITOR_SUPPORT
 /**
  * clutter_wayland_set_compositor_display:
  * @display: A compositor side struct wl_display pointer
@@ -1404,3 +1427,36 @@ clutter_set_windowing_backend (const char *backend_type)
 
   allowed_backend = g_intern_string (backend_type);
 }
+
+PangoDirection
+_clutter_backend_get_keymap_direction (ClutterBackend *backend)
+{
+  ClutterBackendClass *klass;
+
+  klass = CLUTTER_BACKEND_GET_CLASS (backend);
+  if (klass->get_keymap_direction != NULL)
+    return klass->get_keymap_direction (backend);
+
+  return PANGO_DIRECTION_NEUTRAL;
+}
+
+void
+_clutter_backend_reset_cogl_framebuffer (ClutterBackend *backend)
+{
+  if (backend->dummy_onscreen == COGL_INVALID_HANDLE)
+    {
+      CoglError *internal_error = NULL;
+
+      backend->dummy_onscreen = cogl_onscreen_new (backend->cogl_context, 1, 1);
+
+      if (!cogl_framebuffer_allocate (COGL_FRAMEBUFFER (backend->dummy_onscreen),
+                                      &internal_error))
+        {
+          g_critical ("Unable to create dummy onscreen: %s", internal_error->message);
+          cogl_error_free (internal_error);
+          return;
+        }
+    }
+
+  cogl_set_framebuffer (COGL_FRAMEBUFFER (backend->dummy_onscreen));
+}
