@@ -36,13 +36,8 @@
  * that can be used to draw. #ClutterCanvas will emit the #ClutterCanvas::draw
  * signal when invalidated using clutter_content_invalidate().
  *
- * <informalexample id="canvas-example">
- *   <programlisting>
- * <xi:include xmlns:xi="http://www.w3.org/2001/XInclude" parse="text" href="../../../../examples/canvas.c">
- *   <xi:fallback>FIXME: MISSING XINCLUDE CONTENT</xi:fallback>
- * </xi:include>
- *   </programlisting>
- * </informalexample>
+ * See [canvas.c](https://git.gnome.org/browse/clutter/tree/examples/canvas.c?h=clutter-1.18)
+ * for an example of how to use #ClutterCanvas.
  *
  * #ClutterCanvas is available since Clutter 1.10.
  */
@@ -58,14 +53,17 @@
 
 #define CLUTTER_ENABLE_EXPERIMENTAL_API
 
+#include "clutter-actor-private.h"
 #include "clutter-backend.h"
 #include "clutter-cairo.h"
 #include "clutter-color.h"
 #include "clutter-content-private.h"
+#include "clutter-debug.h"
 #include "clutter-marshal.h"
 #include "clutter-paint-node.h"
 #include "clutter-paint-nodes.h"
 #include "clutter-private.h"
+#include "clutter-settings.h"
 
 struct _ClutterCanvasPrivate
 {
@@ -74,7 +72,13 @@ struct _ClutterCanvasPrivate
   int width;
   int height;
 
+  CoglTexture *texture;
+  gboolean dirty;
+
   CoglBitmap *buffer;
+
+  int scale_factor;
+  guint scale_factor_set : 1;
 };
 
 enum
@@ -83,6 +87,8 @@ enum
 
   PROP_WIDTH,
   PROP_HEIGHT,
+  PROP_SCALE_FACTOR,
+  PROP_SCALE_FACTOR_SET,
 
   LAST_PROP
 };
@@ -138,6 +144,8 @@ clutter_canvas_finalize (GObject *gobject)
       priv->buffer = NULL;
     }
 
+  g_clear_pointer (&priv->texture, cogl_object_unref);
+
   G_OBJECT_CLASS (clutter_canvas_parent_class)->finalize (gobject);
 }
 
@@ -177,6 +185,11 @@ clutter_canvas_set_property (GObject      *gobject,
       }
       break;
 
+    case PROP_SCALE_FACTOR:
+      clutter_canvas_set_scale_factor (CLUTTER_CANVAS (gobject),
+                                       g_value_get_int (value));
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
       break;
@@ -201,6 +214,17 @@ clutter_canvas_get_property (GObject    *gobject,
       g_value_set_int (value, priv->height);
       break;
 
+    case PROP_SCALE_FACTOR:
+      if (priv->scale_factor_set)
+        g_value_set_int (value, priv->scale_factor);
+      else
+        g_value_set_int (value, -1);
+      break;
+
+    case PROP_SCALE_FACTOR_SET:
+      g_value_set_boolean (value, priv->scale_factor_set);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
       break;
@@ -244,6 +268,47 @@ clutter_canvas_class_init (ClutterCanvasClass *klass)
                       G_PARAM_READWRITE |
                       G_PARAM_STATIC_STRINGS);
 
+  /**
+   * ClutterCanvas:scale-factor-set:
+   *
+   * Whether the #ClutterCanvas:scale-factor property is set.
+   *
+   * If the #ClutterCanvas:scale-factor-set property is %FALSE
+   * then #ClutterCanvas will use the #ClutterSettings:window-scaling-factor
+   * property.
+   *
+   * Since: 1.18
+   */
+  obj_props[PROP_SCALE_FACTOR_SET] =
+    g_param_spec_boolean ("scale-factor-set",
+                          P_("Scale Factor Set"),
+                          P_("Whether the scale-factor property is set"),
+                          FALSE,
+                          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * ClutterCanvas:scale-factor:
+   *
+   * The scaling factor to be applied to the Cairo surface used for
+   * drawing.
+   *
+   * If #ClutterCanvas:scale-factor is set to a negative value, the
+   * value of the #ClutterSettings:window-scaling-factor property is
+   * used instead.
+   *
+   * Use #ClutterCanvas:scale-factor-set to check if the scale factor
+   * is set.
+   *
+   * Since: 1.18
+   */
+  obj_props[PROP_SCALE_FACTOR] =
+    g_param_spec_int ("scale-factor",
+                      P_("Scale Factor"),
+                      P_("The scaling factor for the surface"),
+                      -1, 1000,
+                      -1,
+                      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+
   /**
    * ClutterCanvas::draw:
    * @canvas: the #ClutterCanvas that emitted the signal
@@ -286,8 +351,10 @@ static void
 clutter_canvas_init (ClutterCanvas *self)
 {
   self->priv = clutter_canvas_get_instance_private (self);
+
   self->priv->width = -1;
   self->priv->height = -1;
+  self->priv->scale_factor = -1;
 }
 
 static void
@@ -296,80 +363,71 @@ clutter_canvas_paint_content (ClutterContent   *content,
                               ClutterPaintNode *root)
 {
   ClutterCanvas *self = CLUTTER_CANVAS (content);
+  ClutterCanvasPrivate *priv = self->priv;
   ClutterPaintNode *node;
-  CoglTexture *texture;
-  ClutterActorBox box;
-  ClutterColor color;
-  guint8 paint_opacity;
-  ClutterScalingFilter min_f, mag_f;
-  ClutterContentRepeat repeat;
 
-  if (self->priv->buffer == NULL)
+  if (priv->buffer == NULL)
     return;
 
-  texture = cogl_texture_new_from_bitmap (self->priv->buffer,
-                                          COGL_TEXTURE_NO_SLICING,
-                                          CLUTTER_CAIRO_FORMAT_ARGB32);
-  if (texture == NULL)
+  if (priv->dirty)
+    g_clear_pointer (&priv->texture, cogl_object_unref);
+
+  if (priv->texture == NULL)
+    priv->texture = cogl_texture_new_from_bitmap (priv->buffer,
+                                                  COGL_TEXTURE_NO_SLICING,
+                                                  CLUTTER_CAIRO_FORMAT_ARGB32);
+
+  if (priv->texture == NULL)
     return;
 
-  clutter_actor_get_content_box (actor, &box);
-  paint_opacity = clutter_actor_get_paint_opacity (actor);
-  clutter_actor_get_content_scaling_filters (actor, &min_f, &mag_f);
-  repeat = clutter_actor_get_content_repeat (actor);
-
-  color.red = paint_opacity;
-  color.green = paint_opacity;
-  color.blue = paint_opacity;
-  color.alpha = paint_opacity;
-
-  node = clutter_texture_node_new (texture, &color, min_f, mag_f);
-  cogl_object_unref (texture);
-
-  clutter_paint_node_set_name (node, "Canvas");
-
-  if (repeat == CLUTTER_REPEAT_NONE)
-    clutter_paint_node_add_rectangle (node, &box);
-  else
-    {
-      float t_w = 1.f, t_h = 1.f;
-
-      if ((repeat & CLUTTER_REPEAT_X_AXIS) != FALSE)
-        t_w = (box.x2 - box.x1) / cogl_texture_get_width (texture);
-
-      if ((repeat & CLUTTER_REPEAT_Y_AXIS) != FALSE)
-        t_h = (box.y2 - box.y1) / cogl_texture_get_height (texture);
-
-      clutter_paint_node_add_texture_rectangle (node, &box,
-                                                0.f, 0.f,
-                                                t_w, t_h);
-    }
-
+  node = clutter_actor_create_texture_paint_node (actor, priv->texture);
+  clutter_paint_node_set_name (node, "Canvas Content");
   clutter_paint_node_add_child (root, node);
   clutter_paint_node_unref (node);
+
+  priv->dirty = FALSE;
 }
 
 static void
 clutter_canvas_emit_draw (ClutterCanvas *self)
 {
   ClutterCanvasPrivate *priv = self->priv;
+  int real_width, real_height;
   cairo_surface_t *surface;
   gboolean mapped_buffer;
   unsigned char *data;
   CoglBuffer *buffer;
+  int window_scale = 1;
   gboolean res;
   cairo_t *cr;
 
   g_assert (priv->width > 0 && priv->width > 0);
 
+  priv->dirty = TRUE;
+
+  if (priv->scale_factor_set)
+    window_scale = priv->scale_factor;
+  else
+    g_object_get (clutter_settings_get_default (),
+                  "window-scaling-factor", &window_scale,
+                  NULL);
+
+  real_width = priv->width * window_scale;
+  real_height = priv->height * window_scale;
+
+  CLUTTER_NOTE (MISC, "Creating Cairo surface with size %d x %d (real: %d x %d, scale: %d)",
+                priv->width, priv->height,
+                real_width, real_height,
+                window_scale);
+
   if (priv->buffer == NULL)
     {
       CoglContext *ctx;
 
       ctx = clutter_backend_get_cogl_context (clutter_get_default_backend ());
       priv->buffer = cogl_bitmap_new_with_size (ctx,
-                                                priv->width,
-                                                priv->height,
+                                                real_width,
+                                                real_height,
                                                 CLUTTER_CAIRO_FORMAT_ARGB32);
     }
 
@@ -389,20 +447,22 @@ clutter_canvas_emit_draw (ClutterCanvas *self)
 
       surface = cairo_image_surface_create_for_data (data,
                                                      CAIRO_FORMAT_ARGB32,
-                                                     priv->width,
-                                                     priv->height,
+                                                     real_width,
+                                                     real_height,
                                                      bitmap_stride);
       mapped_buffer = TRUE;
     }
   else
     {
       surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
-                                            priv->width,
-                                            priv->height);
+                                            real_width,
+                                            real_height);
 
       mapped_buffer = FALSE;
     }
 
+  cairo_surface_set_device_scale (surface, window_scale, window_scale);
+
   self->priv->cr = cr = cairo_create (surface);
 
   g_signal_emit (self, canvas_signals[DRAW], 0,
@@ -576,3 +636,81 @@ clutter_canvas_set_size (ClutterCanvas *canvas,
 
   return clutter_canvas_invalidate_internal (canvas, width, height);
 }
+
+/**
+ * clutter_canvas_set_scale_factor:
+ * @canvas: a #ClutterCanvas
+ * @scale: the scale factor, or -1 for the default
+ *
+ * Sets the scaling factor for the Cairo surface used by @canvas.
+ *
+ * This function should rarely be used.
+ *
+ * The default scaling factor of a #ClutterCanvas content uses the
+ * #ClutterSettings:window-scaling-factor property, which is set by
+ * the windowing system. By using this function it is possible to
+ * override that setting.
+ *
+ * Changing the scale factor will invalidate the @canvas.
+ *
+ * Since: 1.18
+ */
+void
+clutter_canvas_set_scale_factor (ClutterCanvas *canvas,
+                                 int            scale)
+{
+  ClutterCanvasPrivate *priv;
+  GObject *obj;
+
+  g_return_if_fail (CLUTTER_IS_CANVAS (canvas));
+  g_return_if_fail (scale != 0);
+
+  priv = canvas->priv;
+
+  if (scale < 0)
+    {
+      if (!priv->scale_factor_set)
+        return;
+
+      priv->scale_factor_set = FALSE;
+      priv->scale_factor = -1;
+    }
+  else
+    {
+      if (priv->scale_factor_set && priv->scale_factor == scale)
+        return;
+
+      priv->scale_factor_set = TRUE;
+      priv->scale_factor = scale;
+    }
+
+  clutter_content_invalidate (CLUTTER_CONTENT (canvas));
+
+  obj = G_OBJECT (canvas);
+
+  g_object_notify_by_pspec (obj, obj_props[PROP_SCALE_FACTOR]);
+  g_object_notify_by_pspec (obj, obj_props[PROP_SCALE_FACTOR_SET]);
+}
+
+/**
+ * clutter_canvas_get_scale_factor:
+ * @canvas: a #ClutterCanvas
+ *
+ * Retrieves the scaling factor of @canvas, as set using
+ * clutter_canvas_set_scale_factor().
+ *
+ * Return value: the scaling factor, or -1 if the @canvas
+ *   uses the default from #ClutterSettings
+ *
+ * Since: 1.18
+ */
+int
+clutter_canvas_get_scale_factor (ClutterCanvas *canvas)
+{
+  g_return_val_if_fail (CLUTTER_IS_CANVAS (canvas), -1);
+
+  if (!canvas->priv->scale_factor_set)
+    return -1;
+
+  return canvas->priv->scale_factor;
+}
