@@ -21,19 +21,20 @@
  * License along with this library. If not, see <http://www.gnu.org/licenses/>.
  */
 
+#ifndef __CLUTTER_EVENT_H__
+#define __CLUTTER_EVENT_H__
+
 #if !defined(__CLUTTER_H_INSIDE__) && !defined(CLUTTER_COMPILATION)
 #error "Only <clutter/clutter.h> can be included directly."
 #endif
 
-#ifndef __CLUTTER_EVENT_H__
-#define __CLUTTER_EVENT_H__
-
 #include <clutter/clutter-types.h>
 #include <clutter/clutter-input-device.h>
 
 G_BEGIN_DECLS
 
 #define CLUTTER_TYPE_EVENT	        (clutter_event_get_type ())
+#define CLUTTER_TYPE_EVENT_SEQUENCE	(clutter_event_sequence_get_type ())
 
 /**
  * CLUTTER_PRIORITY_EVENTS:
@@ -114,6 +115,8 @@ typedef struct _ClutterScrollEvent      ClutterScrollEvent;
 typedef struct _ClutterStageStateEvent  ClutterStageStateEvent;
 typedef struct _ClutterCrossingEvent    ClutterCrossingEvent;
 typedef struct _ClutterTouchEvent       ClutterTouchEvent;
+typedef struct _ClutterTouchpadPinchEvent ClutterTouchpadPinchEvent;
+typedef struct _ClutterTouchpadSwipeEvent ClutterTouchpadSwipeEvent;
 
 /**
  * ClutterAnyEvent:
@@ -383,6 +386,84 @@ struct _ClutterTouchEvent
   ClutterInputDevice *device;
 };
 
+/**
+ * ClutterTouchpadPinchEvent:
+ * @type: event type
+ * @time: event time
+ * @flags: event flags
+ * @stage: event source stage
+ * @source: event source actor (unused)
+ * @phase: the current phase of the gesture
+ * @x: the X coordinate of the pointer, relative to the stage
+ * @y: the Y coordinate of the pointer, relative to the stage
+ * @dx: movement delta of the pinch focal point in the X axis
+ * @dy: movement delta of the pinch focal point in the Y axis
+ * @angle_delta: angle delta in degrees, clockwise rotations are
+ *   represented by positive deltas
+ * @scale: the current scale
+ *
+ * Used for touchpad pinch gesture events. The current state of the
+ * gesture will be determined by the @phase field.
+ *
+ * Each event with phase %CLUTTER_TOUCHPAD_GESTURE_PHASE_BEGIN
+ * will report a @scale of 1.0, all later phases in the gesture
+ * report the current scale relative to the initial 1.0 value
+ * (eg. 0.5 being half the size, 2.0 twice as big).
+ *
+ * Since: 1.24
+ */
+struct _ClutterTouchpadPinchEvent
+{
+  ClutterEventType type;
+  guint32 time;
+  ClutterEventFlags flags;
+  ClutterStage *stage;
+  ClutterActor *source;
+
+  ClutterTouchpadGesturePhase phase;
+  gfloat x;
+  gfloat y;
+  gfloat dx;
+  gfloat dy;
+  gfloat angle_delta;
+  gfloat scale;
+};
+
+/**
+ * ClutterTouchpadSwipeEvent
+ * @type: event type
+ * @time: event time
+ * @flags: event flags
+ * @stage: event source stage
+ * @source: event source actor (unused)
+ * @phase: the current phase of the gesture
+ * @n_fingers: the number of fingers triggering the swipe
+ * @x: the X coordinate of the pointer, relative to the stage
+ * @y: the Y coordinate of the pointer, relative to the stage
+ * @dx: movement delta of the pinch focal point in the X axis
+ * @dy: movement delta of the pinch focal point in the Y axis
+ *
+ * Used for touchpad swipe gesture events. The current state of the
+ * gesture will be determined by the @phase field.
+ *
+ * Since: 1.24
+ */
+struct _ClutterTouchpadSwipeEvent
+{
+  ClutterEventType type;
+  guint32 time;
+  ClutterEventFlags flags;
+  ClutterStage *stage;
+  ClutterActor *source;
+
+  ClutterTouchpadGesturePhase phase;
+  guint n_fingers;
+  gfloat x;
+  gfloat y;
+  gfloat dx;
+  gfloat dy;
+};
+
 /**
  * ClutterEvent:
  *
@@ -403,28 +484,74 @@ union _ClutterEvent
   ClutterStageStateEvent stage_state;
   ClutterCrossingEvent crossing;
   ClutterTouchEvent touch;
+  ClutterTouchpadPinchEvent touchpad_pinch;
+  ClutterTouchpadSwipeEvent touchpad_swipe;
 };
 
+/**
+ * ClutterEventFilterFunc:
+ * @event: the event that is going to be emitted
+ * @user_data: the data pointer passed to clutter_event_add_filter()
+ *
+ * A function pointer type used by event filters that are added with
+ * clutter_event_add_filter().
+ *
+ * Return value: %CLUTTER_EVENT_STOP to indicate that the event
+ *   has been handled or %CLUTTER_EVENT_PROPAGATE otherwise.
+ *   Returning %CLUTTER_EVENT_STOP skips any further filter
+ *   functions and prevents the signal emission for the event.
+ *
+ * Since: 1.18
+ */
+typedef gboolean (* ClutterEventFilterFunc) (const ClutterEvent *event,
+                                             gpointer            user_data);
+
+CLUTTER_AVAILABLE_IN_ALL
 GType clutter_event_get_type (void) G_GNUC_CONST;
 
+CLUTTER_AVAILABLE_IN_1_20
+GType clutter_event_sequence_get_type (void) G_GNUC_CONST;
+
+CLUTTER_AVAILABLE_IN_ALL
 gboolean                clutter_events_pending                  (void);
+CLUTTER_AVAILABLE_IN_ALL
 ClutterEvent *          clutter_event_get                       (void);
+CLUTTER_AVAILABLE_IN_ALL
 ClutterEvent *          clutter_event_peek                      (void);
+CLUTTER_AVAILABLE_IN_ALL
 void                    clutter_event_put                       (const ClutterEvent     *event);
 
+CLUTTER_AVAILABLE_IN_1_18
+guint                   clutter_event_add_filter                (ClutterStage          *stage,
+                                                                 ClutterEventFilterFunc func,
+                                                                 GDestroyNotify         notify,
+                                                                 gpointer               user_data);
+CLUTTER_AVAILABLE_IN_1_18
+void                    clutter_event_remove_filter             (guint                  id);
+
+CLUTTER_AVAILABLE_IN_ALL
 ClutterEvent *          clutter_event_new                       (ClutterEventType        type);
+CLUTTER_AVAILABLE_IN_ALL
 ClutterEvent *          clutter_event_copy                      (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_ALL
 void                    clutter_event_free                      (ClutterEvent           *event);
 
+CLUTTER_AVAILABLE_IN_ALL
 ClutterEventType        clutter_event_type                      (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_flags                 (ClutterEvent           *event,
                                                                  ClutterEventFlags       flags);
+CLUTTER_AVAILABLE_IN_1_0
 ClutterEventFlags       clutter_event_get_flags                 (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_time                  (ClutterEvent           *event,
                                                                  guint32                 time_);
+CLUTTER_AVAILABLE_IN_ALL
 guint32                 clutter_event_get_time                  (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_state                 (ClutterEvent           *event,
                                                                  ClutterModifierType     state);
+CLUTTER_AVAILABLE_IN_ALL
 ClutterModifierType     clutter_event_get_state                 (const ClutterEvent     *event);
 CLUTTER_AVAILABLE_IN_1_16
 void                    clutter_event_get_state_full            (const ClutterEvent     *event,
@@ -433,26 +560,36 @@ void                    clutter_event_get_state_full            (const ClutterEv
 								 ClutterModifierType    *latched_state,
 								 ClutterModifierType    *locked_state,
 								 ClutterModifierType    *effective_state);
+CLUTTER_AVAILABLE_IN_1_6
 void                    clutter_event_set_device                (ClutterEvent           *event,
                                                                  ClutterInputDevice     *device);
+CLUTTER_AVAILABLE_IN_1_6
 ClutterInputDevice *    clutter_event_get_device                (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_6
 void                    clutter_event_set_source_device         (ClutterEvent           *event,
                                                                  ClutterInputDevice     *device);
 
+CLUTTER_AVAILABLE_IN_1_6
 ClutterInputDevice *    clutter_event_get_source_device         (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_source                (ClutterEvent           *event,
                                                                  ClutterActor           *actor);
+CLUTTER_AVAILABLE_IN_ALL
 ClutterActor *          clutter_event_get_source                (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_stage                 (ClutterEvent           *event,
                                                                  ClutterStage           *stage);
+CLUTTER_AVAILABLE_IN_ALL
 ClutterStage *          clutter_event_get_stage                 (const ClutterEvent     *event);
-
+CLUTTER_AVAILABLE_IN_ALL
 gint                    clutter_event_get_device_id             (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_0
 ClutterInputDeviceType  clutter_event_get_device_type           (const ClutterEvent     *event);
-
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_coords                (ClutterEvent           *event,
                                                                  gfloat                  x,
                                                                  gfloat                  y);
+CLUTTER_AVAILABLE_IN_ALL
 void                    clutter_event_get_coords                (const ClutterEvent     *event,
                                                                  gfloat                 *x,
                                                                  gfloat                 *y);
@@ -465,38 +602,46 @@ float                   clutter_event_get_distance              (const ClutterEv
 CLUTTER_AVAILABLE_IN_1_12
 double                  clutter_event_get_angle                 (const ClutterEvent     *source,
                                                                  const ClutterEvent     *target);
-
+CLUTTER_AVAILABLE_IN_1_6
 gdouble *               clutter_event_get_axes                  (const ClutterEvent     *event,
                                                                  guint                  *n_axes);
-
 CLUTTER_AVAILABLE_IN_1_12
 gboolean                clutter_event_has_shift_modifier        (const ClutterEvent     *event);
 CLUTTER_AVAILABLE_IN_1_12
 gboolean                clutter_event_has_control_modifier      (const ClutterEvent     *event);
 CLUTTER_AVAILABLE_IN_1_12
 gboolean                clutter_event_is_pointer_emulated       (const ClutterEvent     *event);
-
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_key_symbol            (ClutterEvent           *event,
                                                                  guint                   key_sym);
+CLUTTER_AVAILABLE_IN_1_0
 guint                   clutter_event_get_key_symbol            (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_key_code              (ClutterEvent           *event,
                                                                  guint16                 key_code);
+CLUTTER_AVAILABLE_IN_1_0
 guint16                 clutter_event_get_key_code              (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_key_unicode           (ClutterEvent           *event,
                                                                  gunichar                key_unicode);
+CLUTTER_AVAILABLE_IN_1_0
 gunichar                clutter_event_get_key_unicode           (const ClutterEvent     *event);
-
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_button                (ClutterEvent           *event,
                                                                  guint32                 button);
+CLUTTER_AVAILABLE_IN_1_0
 guint32                 clutter_event_get_button                (const ClutterEvent     *event);
+CLUTTER_AVAILABLE_IN_1_0
 guint                   clutter_event_get_click_count           (const ClutterEvent     *event);
-
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_related               (ClutterEvent           *event,
                                                                  ClutterActor           *actor);
+CLUTTER_AVAILABLE_IN_1_0
 ClutterActor *          clutter_event_get_related               (const ClutterEvent     *event);
-
+CLUTTER_AVAILABLE_IN_1_8
 void                    clutter_event_set_scroll_direction      (ClutterEvent           *event,
                                                                  ClutterScrollDirection  direction);
+CLUTTER_AVAILABLE_IN_1_0
 ClutterScrollDirection  clutter_event_get_scroll_direction      (const ClutterEvent     *event);
 CLUTTER_AVAILABLE_IN_1_10
 void                    clutter_event_set_scroll_delta          (ClutterEvent           *event,
@@ -510,13 +655,33 @@ void                    clutter_event_get_scroll_delta          (const ClutterEv
 CLUTTER_AVAILABLE_IN_1_10
 ClutterEventSequence *  clutter_event_get_event_sequence        (const ClutterEvent     *event);
 
+CLUTTER_AVAILABLE_IN_ALL
 guint32                 clutter_keysym_to_unicode               (guint                   keyval);
 CLUTTER_AVAILABLE_IN_1_10
 guint                   clutter_unicode_to_keysym               (guint32                 wc);
 
+CLUTTER_AVAILABLE_IN_1_0
 guint32                 clutter_get_current_event_time          (void);
+CLUTTER_AVAILABLE_IN_1_2
 const ClutterEvent *    clutter_get_current_event               (void);
 
+CLUTTER_AVAILABLE_IN_1_24
+guint                   clutter_event_get_gesture_swipe_finger_count (const ClutterEvent     *event);
+
+CLUTTER_AVAILABLE_IN_1_24
+gdouble                 clutter_event_get_gesture_pinch_angle_delta  (const ClutterEvent     *event);
+
+CLUTTER_AVAILABLE_IN_1_24
+gdouble                 clutter_event_get_gesture_pinch_scale        (const ClutterEvent     *event);
+
+CLUTTER_AVAILABLE_IN_1_24
+ClutterTouchpadGesturePhase clutter_event_get_gesture_phase          (const ClutterEvent     *event);
+
+CLUTTER_AVAILABLE_IN_1_24
+void                    clutter_event_get_gesture_motion_delta       (const ClutterEvent     *event,
+                                                                      gdouble                *dx,
+                                                                      gdouble                *dy);
+
 G_END_DECLS
 
 #endif /* __CLUTTER_EVENT_H__ */
