@@ -74,7 +74,7 @@ G_DEFINE_TYPE_WITH_CODE (ClutterStageGdk,
                          G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_STAGE_WINDOW,
                                                 clutter_stage_window_iface_init));
 
-#ifdef CLUTTER_WINDOWING_X11
+#if defined(GDK_WINDOWING_X11) && defined(COGL_HAS_XLIB_SUPPORT)
 static void
 clutter_stage_gdk_update_foreign_event_mask (CoglOnscreen *onscreen,
 					     guint32 event_mask,
@@ -122,16 +122,21 @@ clutter_stage_gdk_get_geometry (ClutterStageWindow    *stage_window,
 {
   ClutterStageGdk *stage_gdk = CLUTTER_STAGE_GDK (stage_window);
 
-  if (stage_gdk->window != NULL)
+  if (!stage_gdk->foreign_window)
     {
-      geometry->width = gdk_window_get_width (stage_gdk->window);
-      geometry->height = gdk_window_get_height (stage_gdk->window);
+      if (stage_gdk->window != NULL)
+        {
+          geometry->width = gdk_window_get_width (stage_gdk->window);
+          geometry->height = gdk_window_get_height (stage_gdk->window);
+        }
+      else
+        {
+          geometry->width = 640;
+          geometry->height = 480;
+        }
     }
   else
-    {
-      geometry->width = 640;
-      geometry->height = 480;
-    }
+    clutter_stage_window_parent_iface->get_geometry (stage_window, geometry);
 }
 
 static void
@@ -153,7 +158,17 @@ clutter_stage_gdk_resize (ClutterStageWindow *stage_window,
 
   CLUTTER_NOTE (BACKEND, "New size received: (%d, %d)", width, height);
 
-  gdk_window_resize (stage_gdk->window, width, height);
+  /* No need to resize foreign windows, it should be handled by the
+   * embedding framework, but on wayland we might need to resize our
+   * own subsurface.
+   */
+  if (!stage_gdk->foreign_window)
+    gdk_window_resize (stage_gdk->window, width, height);
+#if defined(GDK_WINDOWING_WAYLAND)
+  else if (GDK_IS_WAYLAND_WINDOW (stage_gdk->window))
+    cogl_wayland_onscreen_resize (CLUTTER_STAGE_COGL (stage_gdk)->onscreen,
+                                  width, height, 0, 0);
+#endif
 }
 
 static void
@@ -167,14 +182,130 @@ clutter_stage_gdk_unrealize (ClutterStageWindow *stage_window)
 			 "clutter-stage-window", NULL);
 
       if (stage_gdk->foreign_window)
-	g_object_unref (stage_gdk->window);
+        {
+          ClutterStageCogl *stage_cogl = CLUTTER_STAGE_COGL (stage_window);
+
+          g_object_unref (stage_gdk->window);
+
+          /* Clutter still uses part of the deprecated stateful API of
+           * Cogl (in particulart cogl_set_framebuffer). It means Cogl
+           * can keep an internal reference to the onscreen object we
+           * rendered to. In the case of foreign window, we want to
+           * avoid this, as we don't know what's going to happen to
+           * that window.
+           *
+           * The following call sets the current Cogl framebuffer to a
+           * dummy 1x1 one if we're unrealizing the current one, so
+           * Cogl doesn't keep any reference to the foreign window.
+           */
+          if (cogl_get_draw_framebuffer () == COGL_FRAMEBUFFER (stage_cogl->onscreen))
+            _clutter_backend_reset_cogl_framebuffer (stage_cogl->backend);
+        }
       else
 	gdk_window_destroy (stage_gdk->window);
 
       stage_gdk->window = NULL;
     }
 
-  return clutter_stage_window_parent_iface->unrealize (stage_window);
+  clutter_stage_window_parent_iface->unrealize (stage_window);
+
+#if defined(GDK_WINDOWING_WAYLAND)
+  g_clear_pointer (&stage_gdk->subsurface, wl_subsurface_destroy);
+  g_clear_pointer (&stage_gdk->clutter_surface, wl_surface_destroy);
+#endif
+}
+
+#if defined(GDK_WINDOWING_WAYLAND)
+static struct wl_surface *
+clutter_stage_gdk_wayland_surface (ClutterStageGdk *stage_gdk)
+{
+  GdkDisplay *display;
+  struct wl_compositor *compositor;
+  struct wl_surface *parent_surface;
+  struct wl_region *input_region;
+  gint x, y;
+
+  if (!stage_gdk->foreign_window ||
+      gdk_window_get_window_type (stage_gdk->window) != GDK_WINDOW_CHILD)
+    return gdk_wayland_window_get_wl_surface (stage_gdk->window);
+
+  if (stage_gdk->clutter_surface)
+    return stage_gdk->clutter_surface;
+
+  /* On Wayland if we render to a foreign window, we setup our own
+   * surface to not render in the same buffers as the embedding
+   * framework.
+   */
+  display = gdk_display_get_default ();
+  compositor = gdk_wayland_display_get_wl_compositor (display);
+  stage_gdk->clutter_surface = wl_compositor_create_surface (compositor);
+
+  /* Since we run inside GDK, we can let the embedding framework
+   * dispatch the events to Clutter. For that to happen we need to
+   * disable input on our surface. */
+  input_region = wl_compositor_create_region (compositor);
+  wl_region_add (input_region, 0, 0, 0, 0);
+  wl_surface_set_input_region (stage_gdk->clutter_surface, input_region);
+  wl_region_destroy (input_region);
+
+  parent_surface = gdk_wayland_window_get_wl_surface (gdk_window_get_toplevel (stage_gdk->window));
+  stage_gdk->subsurface = wl_subcompositor_get_subsurface (stage_gdk->subcompositor,
+                                                           stage_gdk->clutter_surface,
+                                                           parent_surface);
+
+  gdk_window_get_origin (stage_gdk->window, &x, &y);
+  wl_subsurface_set_position (stage_gdk->subsurface, x, y);
+  wl_subsurface_set_desync (stage_gdk->subsurface);
+
+  return stage_gdk->clutter_surface;
+}
+#endif
+
+void
+_clutter_stage_gdk_notify_configure (ClutterStageGdk *stage_gdk,
+                                     gint x,
+                                     gint y,
+                                     gint width,
+                                     gint height)
+{
+  if (x < 0 || y < 0 || width < 1 || height < 1)
+    return;
+  if (stage_gdk->foreign_window)
+    {
+#if defined(GDK_WINDOWING_WAYLAND)
+      if (GDK_IS_WAYLAND_WINDOW (stage_gdk->window) &&
+          gdk_window_get_window_type (stage_gdk->window) == GDK_WINDOW_CHILD &&
+          stage_gdk->subsurface)
+        {
+          gint rx, ry;
+          gdk_window_get_origin (stage_gdk->window, &rx, &ry);
+          /* TODO: we might need to apply the scale factor here. */
+          wl_subsurface_set_position (stage_gdk->subsurface, rx, ry);
+        }
+      else
+#endif
+#if defined(GDK_WINDOWING_X11)
+      if (GDK_IS_X11_WINDOW (stage_gdk->window))
+        {
+          ClutterStageCogl *stage_cogl = CLUTTER_STAGE_COGL (stage_gdk);
+          ClutterBackend *backend = CLUTTER_BACKEND (stage_cogl->backend);
+          int scale_factor = gdk_window_get_scale_factor (stage_gdk->window);
+          XConfigureEvent xevent = { ConfigureNotify };
+          xevent.window = GDK_WINDOW_XID (stage_gdk->window);
+          xevent.width = width * scale_factor;
+          xevent.height = height * scale_factor;
+
+          /* Ensure cogl knows about the new size immediately, as we will
+           * draw before we get the ConfigureNotify response. */
+          cogl_xlib_renderer_handle_event (backend->cogl_renderer, (XEvent *)&xevent);
+        }
+      else
+#endif
+        {
+          /* Currently we only support X11 and Wayland. */
+          g_assert_not_reached();
+        }
+    }
 }
 
 static gboolean
@@ -282,6 +413,7 @@ clutter_stage_gdk_realize (ClutterStageWindow *stage_window)
         }
 
       clutter_stage_gdk_set_gdk_geometry (stage_gdk);
+      gdk_window_ensure_native (stage_gdk->window);
     }
   else
     {
@@ -289,8 +421,6 @@ clutter_stage_gdk_realize (ClutterStageWindow *stage_window)
       height = gdk_window_get_height (stage_gdk->window);
     }
 
-  gdk_window_ensure_native (stage_gdk->window);
-
   g_object_set_data (G_OBJECT (stage_gdk->window), "clutter-stage-window", stage_gdk);
 
   stage_cogl->onscreen = cogl_onscreen_new (backend->cogl_context,
@@ -310,7 +440,7 @@ clutter_stage_gdk_realize (ClutterStageWindow *stage_window)
   if (GDK_IS_WAYLAND_WINDOW (stage_gdk->window))
     {
       cogl_wayland_onscreen_set_foreign_surface (stage_cogl->onscreen,
-                                                 gdk_wayland_window_get_wl_surface (stage_gdk->window));
+                                                 clutter_stage_gdk_wayland_surface (stage_gdk));
     }
   else
 #endif
@@ -350,7 +480,7 @@ clutter_stage_gdk_set_fullscreen (ClutterStageWindow *stage_window,
   if (stage == NULL || CLUTTER_ACTOR_IN_DESTRUCTION (stage))
     return;
 
-  if (stage_gdk->window == NULL)
+  if (stage_gdk->window == NULL || stage_gdk->foreign_window)
     return;
 
   CLUTTER_NOTE (BACKEND, "%ssetting fullscreen", is_fullscreen ? "" : "un");
@@ -393,7 +523,7 @@ clutter_stage_gdk_set_title (ClutterStageWindow *stage_window,
 {
   ClutterStageGdk *stage_gdk = CLUTTER_STAGE_GDK (stage_window);
 
-  if (stage_gdk->window == NULL)
+  if (stage_gdk->window == NULL || stage_gdk->foreign_window)
     return;
 
   gdk_window_set_title (stage_gdk->window, title);
@@ -406,7 +536,7 @@ clutter_stage_gdk_set_user_resizable (ClutterStageWindow *stage_window,
   ClutterStageGdk *stage_gdk = CLUTTER_STAGE_GDK (stage_window);
   GdkWMFunction function;
 
-  if (stage_gdk->window == NULL)
+  if (stage_gdk->window == NULL || stage_gdk->foreign_window)
     return;
 
   function = GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE;
@@ -424,7 +554,7 @@ clutter_stage_gdk_set_accept_focus (ClutterStageWindow *stage_window,
 {
   ClutterStageGdk *stage_gdk = CLUTTER_STAGE_GDK (stage_window);
 
-  if (stage_gdk->window == NULL)
+  if (stage_gdk->window == NULL || stage_gdk->foreign_window)
     return;
 
   gdk_window_set_accept_focus (stage_gdk->window, accept_focus);
@@ -440,10 +570,14 @@ clutter_stage_gdk_show (ClutterStageWindow *stage_window,
 
   clutter_actor_map (CLUTTER_ACTOR (CLUTTER_STAGE_COGL (stage_gdk)->wrapper));
 
-  if (do_raise)
-    gdk_window_show (stage_gdk->window);
-  else
-    gdk_window_show_unraised (stage_gdk->window);
+  /* Foreign window should be shown by the embedding framework. */
+  if (!stage_gdk->foreign_window)
+    {
+      if (do_raise)
+        gdk_window_show (stage_gdk->window);
+      else
+        gdk_window_show_unraised (stage_gdk->window);
+    }
 }
 
 static void
@@ -454,7 +588,10 @@ clutter_stage_gdk_hide (ClutterStageWindow *stage_window)
   g_return_if_fail (stage_gdk->window != NULL);
 
   clutter_actor_unmap (CLUTTER_ACTOR (CLUTTER_STAGE_COGL (stage_gdk)->wrapper));
-  gdk_window_hide (stage_gdk->window);
+
+  /* Foreign window should be hidden by the embedding framework. */
+  if (!stage_gdk->foreign_window)
+    gdk_window_hide (stage_gdk->window);
 }
 
 static gboolean
@@ -508,7 +645,7 @@ clutter_stage_gdk_schedule_update (ClutterStageWindow *stage_window,
       return;
     }
 
-  gdk_frame_clock_request_phase (clock, GDK_FRAME_CLOCK_PHASE_UPDATE);
+  gdk_frame_clock_request_phase (clock, GDK_FRAME_CLOCK_PHASE_PAINT);
 
   clutter_stage_window_parent_iface->schedule_update (stage_window, sync_delay);
 }
@@ -563,10 +700,57 @@ clutter_stage_gdk_class_init (ClutterStageGdkClass *klass)
   gobject_class->dispose = clutter_stage_gdk_dispose;
 }
 
+#if defined(GDK_WINDOWING_WAYLAND)
+static void
+registry_handle_global (void *data,
+                        struct wl_registry *registry,
+                        uint32_t name,
+                        const char *interface,
+                        uint32_t version)
+{
+  ClutterStageGdk *stage_gdk = data;
+
+  if (strcmp (interface, "wl_subcompositor") == 0)
+    {
+      stage_gdk->subcompositor = wl_registry_bind (registry,
+                                                   name,
+                                                   &wl_subcompositor_interface,
+                                                   1);
+    }
+}
+
+static void
+registry_handle_global_remove (void *data,
+                               struct wl_registry *registry,
+                               uint32_t name)
+{
+}
+
+static const struct wl_registry_listener registry_listener = {
+  registry_handle_global,
+  registry_handle_global_remove
+};
+#endif
+
 static void
 clutter_stage_gdk_init (ClutterStageGdk *stage)
 {
-  /* nothing to do here */
+#if defined(GDK_WINDOWING_WAYLAND)
+  {
+    GdkDisplay *gdk_display = gdk_display_get_default ();
+    if (GDK_IS_WAYLAND_DISPLAY (gdk_display))
+      {
+        struct wl_display *display;
+        struct wl_registry *registry;
+
+        display = gdk_wayland_display_get_wl_display (gdk_display);
+        registry = wl_display_get_registry (display);
+        wl_registry_add_listener (registry, &registry_listener, stage);
+
+        wl_display_roundtrip (display);
+      }
+  }
+#endif
 }
 
 static void
