@@ -222,6 +222,72 @@ is_touch_device (XIAnyClassInfo         **classes,
   return FALSE;
 }
 
+static gboolean
+is_touchpad_device (ClutterBackendX11 *backend_x11,
+                    XIDeviceInfo      *info)
+{
+  gulong nitems, bytes_after;
+  guint32 *data = NULL;
+  int rc, format;
+  Atom type;
+  Atom prop;
+
+  prop = XInternAtom (backend_x11->xdpy, "libinput Tapping Enabled", True);
+  if (prop == None)
+    return FALSE;
+
+  clutter_x11_trap_x_errors ();
+  rc = XIGetProperty (backend_x11->xdpy,
+                      info->deviceid,
+                      prop,
+                      0, 1, False, XA_INTEGER, &type, &format, &nitems, &bytes_after,
+                      (guchar **) &data);
+  clutter_x11_untrap_x_errors ();
+
+  /* We don't care about the data */
+  XFree (data);
+
+  if (rc != Success || type != XA_INTEGER || format != 8 || nitems != 1)
+    return FALSE;
+
+  return TRUE;
+}
+
+static gboolean
+get_device_ids (ClutterBackendX11  *backend_x11,
+                XIDeviceInfo       *info,
+                gchar             **vendor_id,
+                gchar             **product_id)
+{
+  gulong nitems, bytes_after;
+  guint32 *data = NULL;
+  int rc, format;
+  Atom type;
+
+  clutter_x11_trap_x_errors ();
+  rc = XIGetProperty (backend_x11->xdpy,
+                      info->deviceid,
+                      XInternAtom (backend_x11->xdpy, "Device Product ID", False),
+                      0, 2, False, XA_INTEGER, &type, &format, &nitems, &bytes_after,
+                      (guchar **) &data);
+  clutter_x11_untrap_x_errors ();
+
+  if (rc != Success || type != XA_INTEGER || format != 32 || nitems != 2)
+    {
+      XFree (data);
+      return FALSE;
+    }
+
+  if (vendor_id)
+    *vendor_id = g_strdup_printf ("%.4x", data[0]);
+  if (product_id)
+    *product_id = g_strdup_printf ("%.4x", data[1]);
+
+  XFree (data);
+
+  return TRUE;
+}
+
 static ClutterInputDevice *
 create_device (ClutterDeviceManagerXI2 *manager_xi2,
                ClutterBackendX11       *backend_x11,
@@ -232,9 +298,16 @@ create_device (ClutterDeviceManagerXI2 *manager_xi2,
   ClutterInputMode mode;
   gboolean is_enabled;
   guint num_touches = 0;
+  gchar *vendor_id = NULL, *product_id = NULL;
 
   if (info->use == XIMasterKeyboard || info->use == XISlaveKeyboard)
-    source = CLUTTER_KEYBOARD_DEVICE;
+    {
+      source = CLUTTER_KEYBOARD_DEVICE;
+    }
+  else if (is_touchpad_device (backend_x11, info))
+    {
+      source = CLUTTER_TOUCHPAD_DEVICE;
+    }
   else if (info->use == XISlavePointer &&
            is_touch_device (info->classes, info->num_classes,
                             &touch_source,
@@ -254,6 +327,8 @@ create_device (ClutterDeviceManagerXI2 *manager_xi2,
         source = CLUTTER_CURSOR_DEVICE;
       else if (strstr (name, "wacom") != NULL || strstr (name, "pen") != NULL)
         source = CLUTTER_PEN_DEVICE;
+      else if (strstr (name, "touchpad") != NULL)
+        source = CLUTTER_TOUCHPAD_DEVICE;
       else
         source = CLUTTER_POINTER_DEVICE;
 
@@ -281,6 +356,10 @@ create_device (ClutterDeviceManagerXI2 *manager_xi2,
       break;
     }
 
+  if (info->use != XIMasterKeyboard &&
+      info->use != XIMasterPointer)
+    get_device_ids (backend_x11, info, &vendor_id, &product_id);
+
   retval = g_object_new (CLUTTER_TYPE_INPUT_DEVICE_XI2,
                          "name", info->name,
                          "id", info->deviceid,
@@ -290,11 +369,15 @@ create_device (ClutterDeviceManagerXI2 *manager_xi2,
                          "device-mode", mode,
                          "backend", backend_x11,
                          "enabled", is_enabled,
+                         "vendor-id", vendor_id,
+                         "product-id", product_id,
                          NULL);
 
   translate_device_classes (backend_x11->xdpy, retval,
                             info->classes,
                             info->num_classes);
+  g_free (vendor_id);
+  g_free (product_id);
 
   CLUTTER_NOTE (BACKEND, "Created device '%s' (id: %d, has-cursor: %s)",
                 info->name,
@@ -319,7 +402,7 @@ add_device (ClutterDeviceManagerXI2 *manager_xi2,
    */
   g_hash_table_replace (manager_xi2->devices_by_id,
                         GINT_TO_POINTER (info->deviceid),
-                        g_object_ref (device));
+                        device);
 
   if (info->use == XIMasterPointer ||
       info->use == XIMasterKeyboard)
@@ -401,7 +484,9 @@ translate_hierarchy_event (ClutterBackendX11       *backend_x11,
 
   for (i = 0; i < ev->num_info; i++)
     {
-      if (ev->info[i].flags & XIDeviceEnabled)
+      if (ev->info[i].flags & XIDeviceEnabled &&
+          !g_hash_table_lookup (manager_xi2->devices_by_id,
+                                GINT_TO_POINTER (ev->info[i].deviceid)))
         {
           XIDeviceInfo *info;
           int n_devices;
@@ -649,8 +734,15 @@ translate_coords (ClutterStageX11 *stage_x11,
                   gfloat          *x_out,
                   gfloat          *y_out)
 {
-  *x_out = CLAMP (event_x, 0, stage_x11->xwin_width)  / stage_x11->scale_factor;
-  *y_out = CLAMP (event_y, 0, stage_x11->xwin_height) / stage_x11->scale_factor;
+  ClutterStageCogl *stage_cogl = CLUTTER_STAGE_COGL (stage_x11);
+  ClutterActor *stage = CLUTTER_ACTOR (stage_cogl->wrapper);
+  gfloat stage_width;
+  gfloat stage_height;
+
+  clutter_actor_get_size (stage, &stage_width, &stage_height);
+
+  *x_out = CLAMP (event_x / stage_x11->scale_factor, 0, stage_width);
+  *y_out = CLAMP (event_y / stage_x11->scale_factor, 0, stage_height);
 }
 
 static gdouble
@@ -1128,6 +1220,14 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
 
 #ifdef HAVE_XINPUT_2_2
     case XI_TouchBegin:
+      {
+        XIDeviceEvent *xev = (XIDeviceEvent *) xi_event;
+        device = g_hash_table_lookup (manager_xi2->devices_by_id,
+                                      GINT_TO_POINTER (xev->deviceid));
+        if (!_clutter_input_device_get_stage (device))
+          _clutter_input_device_set_stage (device, stage);
+      }
+      /* Fall through */
     case XI_TouchEnd:
       {
         XIDeviceEvent *xev = (XIDeviceEvent *) xi_event;
@@ -1354,25 +1454,29 @@ clutter_device_manager_xi2_get_core_device (ClutterDeviceManager   *manager,
                                             ClutterInputDeviceType  device_type)
 {
   ClutterDeviceManagerXI2 *manager_xi2 = CLUTTER_DEVICE_MANAGER_XI2 (manager);
-  ClutterBackendX11 *backend_x11;
-  ClutterInputDevice *device;
-  int device_id;
+  ClutterInputDevice *pointer = NULL;
+  GList *l;
 
-  backend_x11 =
-    CLUTTER_BACKEND_X11 (_clutter_device_manager_get_backend (manager));
+  for (l = manager_xi2->master_devices; l != NULL ; l = l->next)
+    {
+      ClutterInputDevice *device = l->data;
+      if (clutter_input_device_get_device_type (device) == CLUTTER_POINTER_DEVICE)
+        {
+          pointer = device;
+          break;
+        }
+    }
 
-  XIGetClientPointer (backend_x11->xdpy, None, &device_id);
-
-  device = g_hash_table_lookup (manager_xi2->devices_by_id,
-                                GINT_TO_POINTER (device_id));
+  if (pointer == NULL)
+    return NULL;
 
   switch (device_type)
     {
     case CLUTTER_POINTER_DEVICE:
-      return device;
+      return pointer;
 
     case CLUTTER_KEYBOARD_DEVICE:
-      return clutter_input_device_get_associated_device (device);
+      return clutter_input_device_get_associated_device (pointer);
 
     default:
       break;
@@ -1435,6 +1539,9 @@ clutter_device_manager_xi2_constructed (GObject *gobject)
     {
       XIDeviceInfo *xi_device = &info[i];
 
+      if (!xi_device->enabled)
+        continue;
+
       add_device (manager_xi2, backend_x11, xi_device, TRUE);
 
       if (xi_device->use == XIMasterPointer ||
@@ -1472,6 +1579,8 @@ clutter_device_manager_xi2_constructed (GObject *gobject)
                                             clutter_x11_get_root_window (),
                                             &event_mask);
 
+  XSync (backend_x11->xdpy, False);
+
   if (G_OBJECT_CLASS (clutter_device_manager_xi2_parent_class)->constructed)
     G_OBJECT_CLASS (clutter_device_manager_xi2_parent_class)->constructed (gobject);
 }
